#!/usr/bin/env python
# -*- coding: UTF-8 -*-

##################################################
# Copyright (C) 2011 Andreas Trawoeger
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
##################################################

from __future__ import division,print_function,absolute_import

"""
Lightdriver for Lightmeter Mark 2.3L using libusb 0.4
"""

import usb
import io
import time
import datetime

VENDOR_ID     = 0x04d8 #: Vendor Id
PRODUCT_ID    = 0xfcb7   #: Product Id

class DeviceDescriptor:
    def __init__(self, vendor_id, product_id, interface_id) :
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.interface_id = interface_id

    def get_device(self) :
        buses = usb.busses()
        for bus in buses :
            for device in bus.devices :
                if device.idVendor == self.vendor_id :
                    if device.idProduct == self.product_id :
                        return device
        return None

class LightDriver(object):
    
    INTERFACE_ID  = 0x0       #: The interface we use to talk to the device
    BULK_IN_EP    = 0x81     #: Endpoint for Bulk reads
    BULK_OUT_EP   = 0x1     #: Endpoint for Bulk writes
    PACKET_LENGTH = 0x40    #: 64 bytes
    TIMEOUT       =  2000
    READ_COUNTS = 'L'                     # Request string for light data
    READ_TEMP = 'T'                      # Request string for temperature data
    READ_ID = 'i'                      # Request string for sensor id
    MASK_COUNTS_STATUS = int('10000000',2)     # Bitmask for light status bit
    MASK_COUNT_DATA =  int('01111111',2)      # Bitmask for light data bits
    MASK_TEMP_STATUS = int('00000111',2)      # Bitmask for temp status bits
    SLEEP = 1.0

    

    @staticmethod 
    def booltostatus(b):
        if b:
            return 'ok'
        else:
            return 'err'
            
    @staticmethod         
    def toint(i):
        if i:
            return int(i)
        else:
            return None
            
    @staticmethod         
    def tofloat(f):
        if f:
            return float(f)
        else:
            return None
            
    @staticmethod         
    def tostr(s):
        if s:
            return str(s)
        else:
            return ''
            
    @staticmethod
    def header():
        return "datetime,station,sensor,temp,counts,status"


    def __init__(self,vendor_id, product_id,interface_id=INTERFACE_ID,
                 debug=False,update=1.0,filename_out=None,buffering=0,
                 station=None,sensor=None,verbose=False,**kwargs):

        self.device_descriptor = DeviceDescriptor(vendor_id,product_id,interface_id)                     
        # The actual device (PyUSB object)
        self.usbdevice = self.device_descriptor.get_device()
        # Handle that is used to communicate with device. Setup in L{open}
        self.usbhandle = None
        self.verbose = verbose
        self.debug = debug
        self.update = self.tofloat(update)
        self.filename_out = self.tostr(filename_out)
        self.buffering = int(buffering)
        self.station = self.tostr(station)
        self.sensor = self.toint(sensor)
        self.temp = None
        self.status = False
        self.counts = None
        self.datetime = None
        
        if self.filename_out:
            self.filehandle=io.open(self.filename_out,'wb',buffering=self.buffering)
            self.filehandle.write(self.header()+'\n')
        else:
            self.filehandle=None
        
        self.open()
        self.readsensor()
                    
    def __str__(self):
        return ("%s,%s,%s,%s,%s,%s"%(self.datetime,self.station,self.sensor,\
                                     self.temp,self.counts,self.booltostatus(self.status)))

    def open(self) :
        self.usbdevice = self.device_descriptor.get_device()
        if not self.usbdevice:
            print ("Lightmeter isn't plugged in")
            self.reset()
        try:
            self.usbhandle = self.usbdevice.open()
            self.usbhandle.claimInterface(self.device_descriptor.interface_id)
        except usb.USBError, err:
            print (err)
            time.sleep(self.SLEEP)
            self.reset()


    def close(self):  
        """ Release device interface """
        try:
            if self.usbhandle:
                self.usbhandle.reset()
                self.usbhandle.releaseInterface()
        except Exception, err:
            print (err)
        self.usbhandle, self.usbdevice = None, None
 
    def write(self, data, timeout=TIMEOUT):
        """
        write(buffer, timeout = 100) -> written

        Write data to printer.
            buffer: data buffer.
            timeout: operation timeout.
        """
        return self.usbhandle.bulkWrite(self.BULK_OUT_EP,data,timeout)

    def read(self, numbytes=PACKET_LENGTH, timeout=TIMEOUT):
        """
        read(numbytes, timeout = 100) -> data

        Read data from printer.
            numbytes: number of bytes to read.
            timeout: operation timeout.
        """
        return self.usbhandle.bulkRead(self.BULK_IN_EP,numbytes,timeout)
    
    def reset(self):
        self.close()
        time.sleep(self.SLEEP)
        self.open()
        
    def getval(self,question):
        """Get value from USB device"""
        try:
            self.write(question)    
            return self.read()
        except usb.core.USBError:
            self.reset()
            return self.getval(question)
            
    def readsensor(self):
        """Read Serial Number"""
        req = self.getval(self.READ_ID)
        
        if req:
            snrlow,snrhigh = req
            if snrhigh <> 255 and snrlow <> 255:
                snr = snrhigh*256 + snrlow

                if snr < 10000:
                    self.sensor = snr + 910000
                else:
                    self.sensor= snr + 1000000
        else:
            self.sensor=None


    def setversion(self,version=None):
        if version:
            self.version=version
        else:
            if not self.sensor or int(self.sensor) < 910000:
                self.version="2.3"
            elif int(self.sensor) > 1000000:
                self.version="2.4"
            else:
                self.version="2.3L"

        
    def readcounts(self):
        """Read counts value"""
        req = self.getval(self.READ_COUNTS)
        
        self.counts=None        
        self.status=False
        
        if self.debug:
            print (len(req),req)
                  
        if len(req) ==7 : 
            countslow, countshigh, valrange, visuallow, visualhigh, infraredlow, infraredhigh = req

            self.status = (countshigh & self.MASK_COUNTS_STATUS) == 0               # Check if status bit == 0
            countsraw = ((countshigh & self.MASK_COUNT_DATA) << 8)  + countslow
            
            if valrange == 1:
                self.counts = countsraw * 120
            elif valrange == 2:
                self.counts = countsraw * 8
            elif valrange == 3:
                self.counts = countsraw * 4
            elif valrange == 4:
                self.counts = countsraw * 2
            elif valrange == 5:
                self.counts = countsraw

     
    def readtemp(self):
        """Read temperature"""
        req = self.getval(self.READ_TEMP)
        
        if len(req) == 2:
            templow, temphigh = req
        
            status = (templow & self.MASK_TEMP_STATUS) == 0            # Check if status bit == 0
            tempraw = ((temphigh << 8)  + templow ) >> 3        # Statusbits ausblenden
            temp = 625.0 * tempraw / 10000.0                    # Umrechnung von 1/16° zu 1/10°
        else:
            temp = -1 
            status = False
        
        self.temp = temp

        
    def reader(self):
        """Continously read from lightmeter"""
        dtupdate =  datetime.timedelta(seconds=self.update)
        dtmes = datetime.datetime.utcnow()

        
        while True:      
            dtnow = datetime.datetime.utcnow()
            dtuncalibrated = dtmes + dtupdate
            dtcalibrated = dtuncalibrated.replace(microsecond=dtuncalibrated.microsecond//100000*100000)
            dtsleep = dtcalibrated - dtnow
            sleep = dtsleep.seconds+(dtsleep.microseconds)/1000.0/1000.0
            if sleep <= self.update:
                time.sleep(sleep)     
            dtmes = datetime.datetime.utcnow()     
            self.datetime = dtmes

            try:  
                self.readtemp()
                self.readcounts()
            except AttributeError:
                print ("No lightmeter found")
                self.reset()
                continue
                
            if self.verbose:
                print (self.__str__())
                
            if self.filehandle:
                self.filehandle.write(self.__str__()+'\n')
       
def main(vendor_id=VENDOR_ID, product_id=PRODUCT_ID):
    """run lightmeter"""
    
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-c","--connect", dest="connect", default=False, action="store_true",
                                      help="connect to lightmeter")
    parser.add_option("-d","--debug", dest="debug", default=False, action="store_true",
                                      help="enable debug")
    parser.add_option("-f","--frequency", dest="update", default=1.0, metavar="seconds",
                                      help="update frequency in seconds [%default]")
    parser.add_option("-n","--station", dest="station", default="", metavar="id",
                                      help="station id")
    parser.add_option("-s","--sensor", dest="sensor", default="", metavar="snr",
                                      help="sensor serial number")
    parser.add_option("-o","--filename-out", dest="filename_out", default='', metavar="file",
                                      help="filename for data logging")
    parser.add_option("-a","--filename-auto", dest="filenameauto", default=False, action="store_true",
                                      help="Automatically generate filename")
    parser.add_option("-B","--buffering", dest="buffering", default=0, metavar="bytes",
                                      help="buffer file I/O [%default]")
    parser.add_option("-v", "--verbose",action="store_true", dest="verbose", default=False,
                                      help="be verbose")
    (options, args) = parser.parse_args()

    if not options.connect:
        parser.print_help()
        exit(-1)
             
    if options.filenameauto:
        dtnow = datetime.datetime.utcnow()
        datenow = "%04i_%02i%02i"%(dtnow.year,dtnow.month,dtnow.day)
        timenow = "%02i%02i%02i"%(dtnow.hour,dtnow.minute,dtnow.second)
        options.filename_out = options.filename_out+options.station+'_'+datenow+'_'+timenow+'.csv'

    lm=LightDriver(vendor_id=vendor_id,product_id=product_id,**vars(options))
    lm.reader()       
           
if __name__ == "__main__":
    main()


